from typing import Optional, Tuple
from logging import Logger

from org.codeforrussia.selector.config.global_config import GlobalConfig
from org.codeforrussia.selector.standardizer.common import Standardizer
from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel, ElectionType, ElectionLocationType
from org.codeforrussia.selector.standardizer.recognizers.protocol_field_recognizer_registry_factory import \
    ProtocolFieldRecognizerRegistryFactory
from org.codeforrussia.selector.standardizer.recognizers.protocol_field_recognizers import ProtocolRow
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import \
    StandardProtocolSchemaRegistryFactory


class ShpilkinDumpStandardizer(Standardizer):
    logger = Logger("ShpilkinDumpStandardizer")

    def __init__(self, schema_registry_factory: StandardProtocolSchemaRegistryFactory,
                 protocol_recognizer_registry_factory: ProtocolFieldRecognizerRegistryFactory,
                 global_config: GlobalConfig,
                 ):
        super().__init__(schema_registry_factory, protocol_recognizer_registry_factory, global_config)

    def convert(self, protocol_data: dict) -> Optional[Tuple[dict, ElectionLevel, ElectionType, ElectionLocationType]]:
        """
        Accepts single-protocol data in Shpilkin's format.
        First, it recognizes the election attributes by election name.
        :param protocol_data: data
        :return: standardized data
        """
        if protocol_data["loc"]:
            election_name = protocol_data["loc"][0]
            election_level, election_type, election_location_type = self.detect_election_attributes_by_name(election_name)

            election_location = protocol_data["loc"][1:]
            election_commission_level = self.detect_commission_level(election_location)
            election_date = "21.09.2020"  # must be included, but no provided in dump; TODO: extract from filename

            sdata = {
                "election": {
                    "name": election_name,
                    "location": election_location,
                    "commission_level": election_commission_level,
                    "date": election_date
                }
            }

            protocol_rows = [ProtocolRow(line_number=d["line_num"],
                             line_name=d["line_name"],
                             line_value=int(d["line_val"])) for d in protocol_data['data']]

            recognized_schema = self.schema_registry.search_schema(election_level, election_type, election_location_type)
            standardized_protocol_rows = self.protocol_recognizer_registry.get_recognizer(election_level).recognize(protocol_rows, recognized_schema)
            sdata = dict(sdata, **standardized_protocol_rows)

            return sdata, election_level, election_type, election_location_type
        else:
            self.logger.warning("Protocol data is likely malformed. Empty result will be returned for this protocol data.")
            return None



import os
from pathlib import Path
import json
from typing import Tuple, Dict, List

from fastavro import parse_schema
from collections import OrderedDict
from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel, ElectionType, ElectionLocationType

class StandardProtocolSchemaRegistryFactory(object):
    def get_schema_registry():
        class StandardProtocolSchemaRegistry:
            """
            Registry of supported standard schemas of election protocols. They cover different levels, election types (personal vs representatives) and regions.
            Specifically,
            - on FEDERAL level there are two elections: Presidential and State Duma elections;
            - on REGIONAL level there are two: elections of Governors and elections of deputies to regional parliaments
            - on MUNICIPAL level there are four:
                - personal: heads of municipal district and heads of city / rural locations
                - representative: representatives at municipal district level and city / rural location level.

            Additionally, we have a common election schema for common attributes.
            """
            def __init__(self):
                named_schemas = {}

                self._registered_schemas: Dict[Tuple[ElectionLevel, ElectionType, ElectionLocationType], Dict]  = OrderedDict() # dict where key is tuple(level, type, location) and value is schema

                schemas_dir = Path(os.path.dirname(__file__))

                with open( str(schemas_dir / "common" / "election_1_0.avsc"), "rb") as election_schema_file:
                    self._election_schema = parse_schema(json.load(election_schema_file), named_schemas)

                with open(str(schemas_dir / "federal" / "state_duma_1_0.avsc"), "rb") as state_duma_schema_file:
                    self._registered_schemas[(ElectionLevel.FEDERAL, ElectionType.REPRESENTATIVE, None)] = parse_schema(json.load(state_duma_schema_file), named_schemas)

                with open(str(schemas_dir / "regional" / "head_1_0.avsc"), "rb") as regional_head_file:
                    self._registered_schemas[(ElectionLevel.REGIONAL, ElectionType.PERSONAL, None)] = parse_schema(json.load(regional_head_file), named_schemas)

                with open(str(schemas_dir / "municipal" / "deputy_city_1_0.avsc"), "rb") as municipal_deputy_city_file:
                    self._registered_schemas[(ElectionLevel.MUNICIPAL, ElectionType.REPRESENTATIVE, ElectionLocationType.CITY_RURAL)] = parse_schema(json.load(municipal_deputy_city_file), named_schemas)

            def get_common_election_schema(self) -> Dict:
                """
                Returns election schema with common attributes for protocols of different kinds
                :return:
                """
                return self._election_schema

            def get_all_registered_schema_keys(self) -> List[Tuple[ElectionLevel, ElectionType, ElectionLocationType]]:
                return list(self._registered_schemas.keys())

            def search_schema(self, level: ElectionLevel, type: ElectionType, location: ElectionLocationType = None) -> Dict:
                """
                Searches the standardized schema by criteria
                :param level: election level
                :param type: election type
                :param location: election location type
                :return: schema as dict
                """
                try:
                    return self._registered_schemas[(level, type, location)]
                except KeyError:
                    raise NotImplementedError(f"This combination of election level = '{level}', type = '{type}', location = '{location}' is not supported yet")

        return StandardProtocolSchemaRegistry()

    get_schema_registry = staticmethod(get_schema_registry)





from google.cloud import storage

class GCS:
    def __init__(self):
        self.client = storage.Client()

    def download_file(self, bucket_name: str, filename: str, output_local_filename: str):
        bucket = self.client.get_bucket(bucket_name)
        blob = bucket.blob(filename)
        blob.download_to_filename(output_local_filename)


from abc import ABC, abstractmethod
from typing import Optional, Tuple, List, Dict

from org.codeforrussia.selector.config.global_config import GlobalConfig
from org.codeforrussia.selector.standardizer.recognizers.election_attribute_recognizer import ElectionAttributeRecognizer
from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel, ElectionLocationType, ElectionType
from org.codeforrussia.selector.standardizer.recognizers.protocol_field_recognizer_registry_factory import \
    ProtocolFieldRecognizerRegistryFactory
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import StandardProtocolSchemaRegistryFactory

class Standardizer(ABC):

    def __init__(self,
                 schema_registry_factory: StandardProtocolSchemaRegistryFactory,
                 protocol_recognizer_registry_factory: ProtocolFieldRecognizerRegistryFactory,
                 global_config: GlobalConfig,
                 ):
        self.schema_registry = schema_registry_factory.get_schema_registry()
        self.protocol_recognizer_registry = protocol_recognizer_registry_factory.get_registry(global_config=global_config)
        self.election_attribute_recognizer = ElectionAttributeRecognizer()
        self.commission_levels = [f for f in self.schema_registry.get_common_election_schema()["fields"] if f["name"] == "commission_level"][0]["type"]["symbols"]

    def detect_election_attributes_by_name(self, election_name) -> Tuple[ElectionLevel, ElectionType, ElectionLocationType]:
        return self.election_attribute_recognizer.recognize(election_name)

    def detect_commission_level(self, election_location: [str]) -> str:
        """
        Given election location, recognizes the commission level according to the levels defined in the common schema
        :param election_location: election location
        :return:
        """
        if "УИК " in election_location[-1]:
            return self.commission_levels[-1]
        elif any(x in election_location[-1] for x in ["ТИК", "округ"]):
            return self.commission_levels[-2]
        elif "ОИК" in election_location[-1]:
            return self.commission_levels[-3]
        else: # Fallback is territorial
            return self.commission_levels[-2]

    def convert_batch(self, batch: List[Dict]) -> List[Optional[Dict]]:
        def tuple_to_dict(output: Optional[Tuple[Dict, ElectionLevel, ElectionType, ElectionLocationType]]) -> Optional[Dict]:
            if output is not None:
                sdata, election_level, election_type, election_location = output
                return {
                    "sdata": sdata,
                    "election_attrs": {"level": election_level, "type": election_type, "location": election_location}
                        }
            else:
                return None

        return [tuple_to_dict(self.convert(protocol_data)) for protocol_data in batch]



    @abstractmethod
    def convert(self, protocol_data: dict) -> Optional[Tuple[dict, ElectionLevel, ElectionType, ElectionLocationType]]:
        pass




from org.codeforrussia.selector.config.global_config import GlobalConfig
from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel
from org.codeforrussia.selector.standardizer.recognizers.protocol_field_recognizers import LineNumberBasedProtocolRecognizer, \
    SimilarityBasedProtocolRecognizer, ProtocolFieldRecognizer


class ProtocolFieldRecognizerRegistryFactory(object):

    def get_registry(global_config: GlobalConfig):
        class ProtocolFieldRecognizerRegistry:
            """
            Registry of supported election levels and their recognizers
            """
            def __init__(self, global_config: GlobalConfig):
                # share the same similarity model
                similarity_based_protocol_recognizer = SimilarityBasedProtocolRecognizer(
                                                            global_config=global_config,                                                                             model_name="similarity-protocol-recognizer_1_3_0",
                )
                self.recognizers = {
                    ElectionLevel.FEDERAL: LineNumberBasedProtocolRecognizer(),
                    ElectionLevel.REGIONAL: similarity_based_protocol_recognizer,
                    ElectionLevel.MUNICIPAL: similarity_based_protocol_recognizer,
                }

            def get_recognizer(self, election_level: ElectionLevel) -> ProtocolFieldRecognizer:
                try:
                    return self.recognizers[election_level]
                except KeyError:
                    raise NotImplementedError(f"This election level is not supported yet: {election_level}")
        return ProtocolFieldRecognizerRegistry(global_config=global_config)

    get_registry = staticmethod(get_registry)


from setuptools import setup, find_packages

setup(
    name='selector-standardizers',
    version='0.7.0',
    description='Electoral Data Standardization classes for the Selector project',
    long_description=open("README.md", encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    author='Nikita Zhiltsov',
    author_email='mail@codeforrussia.org',
    url='https://github.com/Code-for-Russia/selector-pipeline',
    packages=find_packages(where="src", exclude='test'),  # same as name
    package_dir={'': 'src'},
    package_data={
        'org.codeforrussia.selector.standardizer.schemas.common': ['*.avsc'],
        'org.codeforrussia.selector.standardizer.schemas.federal': ['*.avsc'],
        'org.codeforrussia.selector.standardizer.schemas.regional': ['*.avsc'],
        'org.codeforrussia.selector.standardizer.schemas.municipal': ['*.avsc'],
                  },
    install_requires=[
        'pytest>=6.2.4',
        'fastavro>=1.4.0',
        'jsonlines>=2.0.0',
        'dataclasses>=0.6',
        'sentence_transformers>=1.2.0',
        'scikit-learn>=0.24.2',
        'google-cloud-storage>=1.38.0',
    ],
    include_package_data=True,
    python_requires='>=3.7'
)